#include "XAudioVorbis.h"
#include <stdio.h>

#pragma warning( disable: 4996 )

#ifdef _DEBUG
#pragma comment( lib, "lib\\vorbis\\win32\\VorbisFile_Static_Debug\\vorbisfile_static_d.lib" )
#pragma comment( lib, "lib\\vorbis\\win32\\Vorbis_Static_Debug\\vorbis_static_d.lib" )
#pragma comment( lib, "lib\\ogg\\win32\\Static_Debug\\ogg_static_d.lib" )
#else
#pragma comment( lib, "lib\\vorbis\\win32\\VorbisFile_Static_Release\\vorbisfile_static.lib" )
#pragma comment( lib, "lib\\vorbis\\win32\\Vorbis_Static_Release\\vorbis_static.lib" )
#pragma comment( lib, "lib\\ogg\\win32\\Static_Release\\ogg_static.lib" )
#endif

XAudioVorbis::XAudioVorbis()
: fp( NULL ), pSourceVoice( NULL ), BufferTopPos( 0 ), IsBufferedAll( false ), NumLoops( 0 ), pXAudioFadeEffect( NULL )
{
	pXAudioFadeEffect = new XAudioFadeEffect();
	pStatusCallBack = new VorbisCallback();

	for( int i = 0; i < WAV_NUMBUFFER; ++i ){
		pBuffer[i] = new char [WAV_BUFFERSIZE];
	}

	memset( &vf, 0, sizeof( OggVorbis_File ) );

	Status = STATUS_WAITING;
}

XAudioVorbis::~XAudioVorbis()
{
	Release();

	if( pStatusCallBack ){
		delete pStatusCallBack;
		pStatusCallBack = NULL;
	}

	if( pXAudioFadeEffect ){
		delete pXAudioFadeEffect;
		pXAudioFadeEffect = NULL;
	}

	for( int i = 0; i < WAV_NUMBUFFER; ++i ){
		delete [] pBuffer[i];
		pBuffer[i] = NULL;
	}
}

bool XAudioVorbis::Release()
{
	Stop();

	if( vf.datasource != 0 ){
		ov_clear( &vf );
		memset( &vf, 0, sizeof( OggVorbis_File ) );
	}

	if( fp ){
		fclose( fp );
		fp = NULL;
	}

	if( pSourceVoice ){
		pSourceVoice->FlushSourceBuffers();

		for( int i = 0; i < 10; ++i ){ // 10 ͓K
	        XAUDIO2_VOICE_STATE s;
			pSourceVoice->GetState( &s );
			if( s.BuffersQueued == 0 ){
				break;
			}

			Sleep( 100 );
		}

        pSourceVoice->DestroyVoice();
		pSourceVoice = NULL;

		ReleaseCallback_BufferEnd( Callback_BufferEnd );
	}

	Status = STATUS_WAITING;

	BufferTopPos = 0;
	IsBufferedAll = false;
	NumLoops = 0;

	return true;
}

bool XAudioVorbis::Load( const char *pFileName )
{
	Release();

	fp = fopen( pFileName, "rb" );
	if( !fp ) return false;

	if( ov_open( fp, &vf, NULL, 0 ) < 0 ){
		fclose( fp );
		return false;
	}

	for( int i = 0; i < WAV_NUMBUFFER; ++i ){

		unsigned int ReadSize;
		Read( pBuffer[i], WAV_BUFFERSIZE, ReadSize, IsBufferedAll );
		if( ReadSize <= 0 || IsBufferedAll == true ){
			break;
		}

		ValidBufferSize[i] = ReadSize;
		BufferTopPos = ( i + 1 ) % WAV_NUMBUFFER;
	}

    vorbis_info *vi = ov_info( &vf, -1 );

	// PCM audio
	WAVEFORMATEX WaveFormatEx;

	WaveFormatEx.wFormatTag = 1; // PCM audio
	WaveFormatEx.nChannels = vi->channels;
	WaveFormatEx.nSamplesPerSec = vi->rate;
	WaveFormatEx.wBitsPerSample = 16;
	WaveFormatEx.nBlockAlign = WaveFormatEx.nChannels * WaveFormatEx.wBitsPerSample / 8;
	WaveFormatEx.nAvgBytesPerSec = WaveFormatEx.nSamplesPerSec * WaveFormatEx.nBlockAlign;
	WaveFormatEx.cbSize = 0;

	pStatusCallBack->pXAudioVorbis = this;

	if( FAILED( XAudio::GetInstance().GetXAudio2()->CreateSourceVoice(
			&pSourceVoice, &WaveFormatEx, 0, XAUDIO2_DEFAULT_FREQ_RATIO, pStatusCallBack ) ) ){
		return false;
	}

	if( !pXAudioFadeEffect->Init( pSourceVoice, vi->channels ) ) return false;

	if( BufferTopPos == 0 ){
		for( int i = 0; i < WAV_NUMBUFFER; ++i ){
			XAUDIO2_BUFFER XAudio2Buffer;
			memset( &XAudio2Buffer, 0, sizeof( XAUDIO2_BUFFER ) );

			XAudio2Buffer.AudioBytes = ValidBufferSize[i];
			XAudio2Buffer.pAudioData = (BYTE*)pBuffer[i];
			XAudio2Buffer.pContext = this;

			if( FAILED( pSourceVoice->SubmitSourceBuffer( &XAudio2Buffer ) ) ){
				return false;
			}
		}
	}
	else{
		for( int i = 0; i < BufferTopPos - 1; ++i ){
			XAUDIO2_BUFFER XAudio2Buffer;
			memset( &XAudio2Buffer, 0, sizeof( XAUDIO2_BUFFER ) );

			XAudio2Buffer.AudioBytes = ValidBufferSize[i];
			XAudio2Buffer.pAudioData = (BYTE*)pBuffer[i];
			XAudio2Buffer.pContext = this;

			if( i == BufferTopPos - 2 ) XAudio2Buffer.Flags = XAUDIO2_END_OF_STREAM;

			if( FAILED( pSourceVoice->SubmitSourceBuffer( &XAudio2Buffer ) ) ){
				return false;
			}
		}
	}

	FadeEffectParams Param;
	Param.IsFade = false;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	SetCallback_BufferEnd( Callback_BufferEnd );

	IsBufferedAll = false;

	return true;
}

bool XAudioVorbis::Play( int NumLoops )
{
	if( pSourceVoice ){
		Stop();
		Fade( 0, 1.0f, 1.0f );

		if( NumLoops >= 0 ){
			this->NumLoops = NumLoops;
		}
		else{
			this->NumLoops = XAUDIOVORBIS_LOOP_INFINITE;
		}

		if( FAILED( pSourceVoice->Start( 0 ) ) ){
			return false;
		}
	}

	Status = STATUS_PLAYING;

	return true;
}

bool XAudioVorbis::Resume()
{
	if( pSourceVoice ){
		if( FAILED( pSourceVoice->Start( 0 ) ) ){
			return false;
		}
	}

	Status = STATUS_PLAYING;

	return true;
}

bool XAudioVorbis::Stop()
{
	if( pSourceVoice ){
		if( FAILED( pSourceVoice->Stop( XAUDIO2_PLAY_TAILS ) ) ){
			return false;
		}
	}

	Status = STATUS_STOP;

	return true;
}

void XAudioVorbis::Read( char *pBuffer, const unsigned int BufferSize, unsigned int &ReadSize, bool &IsEnd )
{
	int CurrentSection;
	char Buffer[4096];

	ReadSize = 0;
	IsEnd = false;

	while( 1 ){
		int Size = ov_read( &vf, Buffer, 4096, 0, 2, 1, &CurrentSection );
		if( Size <= 0 ){
			IsEnd = true;
			break;
		}

		memcpy( pBuffer + ReadSize, Buffer, Size );

		ReadSize += Size;
		if( ReadSize + 4096 >= BufferSize ) break;
	}
}

// -2^24 to 2^24
// 1.0 means there is no attenuation or gain and 0 means silence. Negative levels can be used to invert the audio's phase.
bool XAudioVorbis::SetVolume( const float Volume )
{
	if( Volume < -16777216.0 || Volume > 16777216.0 ) return false;

	if( FAILED( pSourceVoice->SetVolume( Volume ) ) ){
		return false;
	}

	return true;
}

float XAudioVorbis::GetVolume() const
{
	float Volume;
	pSourceVoice->GetVolume( &Volume );

	return Volume;
}

bool XAudioVorbis::SetFrequencyRatio( const float Ratio )
{
	if( Ratio < XAUDIO2_MIN_FREQ_RATIO || Ratio > XAUDIO2_DEFAULT_FREQ_RATIO ) return false;

	if( FAILED( pSourceVoice->SetFrequencyRatio( Ratio ) ) ){
		return false;
	}

	return true;
}

float XAudioVorbis::GetFrequencyRatio() const
{
	float Ratio;
	pSourceVoice->GetFrequencyRatio( &Ratio );

	return Ratio;
}

bool XAudioVorbis::Fade( const unsigned int Milliseconds, const float BeginGain, const float EndGain )
{
	FadeEffectParams Param;
	Param.IsFade = true;
	Param.BeginGain = BeginGain;
	Param.EndGain = EndGain;
	Param.Milliseconds = Milliseconds;
	Param.Frames = 0;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	return true;
}

// PrevParam.Gain̏KvȂ̂ŁAKXAudioWave::Fade( unsigned int Milliseconds, const float BeginGain, const float EndGain )Ă񂾌ŎgƁB
bool XAudioVorbis::Fade( const unsigned int Milliseconds, const float EndGain )
{
	FadeEffectParams PrevParam;
	if( FAILED( pSourceVoice->GetEffectParameters( 0, &PrevParam, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	FadeEffectParams Param;
	Param.IsFade = true;
	Param.BeginGain = PrevParam.Gain;
	Param.EndGain = EndGain;
	Param.Milliseconds = Milliseconds;
	Param.Frames = 0;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	return true;
}

void XAudioVorbis::Callback_BufferEnd( void *p )
{
	XAudioVorbis *pXAudioVorbis = (XAudioVorbis *)p;

	if( pXAudioVorbis->IsBufferedAll ) return;

	unsigned int ReadSize;
	pXAudioVorbis->Read( pXAudioVorbis->pBuffer[pXAudioVorbis->BufferTopPos], WAV_BUFFERSIZE, ReadSize, pXAudioVorbis->IsBufferedAll );

	pXAudioVorbis->ValidBufferSize[pXAudioVorbis->BufferTopPos] = ReadSize;

	XAUDIO2_BUFFER XAudio2Buffer;
	memset( &XAudio2Buffer, 0, sizeof( XAUDIO2_BUFFER ) );

	XAudio2Buffer.AudioBytes = pXAudioVorbis->ValidBufferSize[pXAudioVorbis->BufferTopPos];
	XAudio2Buffer.pAudioData = (BYTE*)pXAudioVorbis->pBuffer[pXAudioVorbis->BufferTopPos];
	XAudio2Buffer.pContext = pXAudioVorbis;

	if( pXAudioVorbis->IsBufferedAll ){
		if( pXAudioVorbis->NumLoops == 0 ){
			XAudio2Buffer.Flags = XAUDIO2_END_OF_STREAM;
		}
		else{
			ov_time_seek( &( pXAudioVorbis->vf ), 0 );
			pXAudioVorbis->IsBufferedAll = false;

			if( pXAudioVorbis->NumLoops != XAUDIOVORBIS_LOOP_INFINITE ){
				--( pXAudioVorbis->NumLoops );
			}
		}
	}

	if( ReadSize > 0 ){
		if( FAILED( pXAudioVorbis->pSourceVoice->SubmitSourceBuffer( &XAudio2Buffer ) ) ){
			return;
		}

		pXAudioVorbis->BufferTopPos = ( pXAudioVorbis->BufferTopPos + 1 ) % WAV_NUMBUFFER;
	}
}
