#ifndef __XAUDIOVORBIS_H__
#define __XAUDIOVORBIS_H__

#include "XAudio.h"
#include "XAudioFadeEffect.h"
#include <vector>
#include <algorithm>
#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

#define WAV_BUFFERSIZE ( 1024 * 32 )
// WAV_NUMBUFFERSubmitSourceBuffer64𒴂Ƃ̂łȉɂ邱
#define WAV_NUMBUFFER 16

class VorbisCallback;

// The vorbis libraries are fully reentrant, but not thread safe.
class XAudioVorbis{

private:
typedef void (*PFUNC)();
typedef void (*PFUNC_VOIDP)( void * );
typedef void (*PFUNC_VOIDP_HRESULT)( void *, HRESULT );

public:
enum STATUS{
	STATUS_WAITING,
	STATUS_PLAYING,
	STATUS_STOP,
};


#define XAUDIOVORBIS_LOOP_INFINITE INT_MIN

public:
	XAudioVorbis();
	~XAudioVorbis();

	bool Load( const char *pFileName );
	bool Release();

	bool Play( int NumLoops = 0 );
	bool Resume();
	bool Stop();
	STATUS GetStatus(){ return Status; };

	bool SetVolume( const float Volume );
	float GetVolume() const;
	bool SetFrequencyRatio( const float Ratio );
	float GetFrequencyRatio() const;

	bool Fade( const unsigned int Milliseconds, const float BeginGain, const float EndGain );
	bool Fade( const unsigned int Milliseconds, const float EndGain );

	static void Callback_BufferEnd( void *p );

	void SetCallback_StreamEnd(                PFUNC               pFunc ){ Callback_StreamEnds.push_back(                pFunc ); };
	void SetCallback_VoiceProcessingPassEnd(   PFUNC               pFunc ){ Callback_VoiceProcessingPassEnds.push_back(   pFunc ); };
	void SetCallback_VoiceProcessingPassStart( PFUNC               pFunc ){ Callback_VoiceProcessingPassStarts.push_back( pFunc ); };
	void SetCallback_BufferEnd(                PFUNC_VOIDP         pFunc ){ Callback_BufferEnds.push_back(                pFunc ); };
	void SetCallback_BufferStart(              PFUNC_VOIDP         pFunc ){ Callback_BufferStarts.push_back(              pFunc ); };
	void SetCallback_LoopEnd(                  PFUNC_VOIDP         pFunc ){ Callback_LoopEnds.push_back(                  pFunc ); };
	void SetCallback_VoiceError(               PFUNC_VOIDP_HRESULT pFunc ){ Callback_VoiceErrors.push_back(               pFunc ); };

	void ReleaseCallback_StreamEnd(                PFUNC               pFunc ){ Callback_StreamEnds.erase(                std::find( Callback_StreamEnds.begin(),                Callback_StreamEnds.end(),                pFunc ) ); };
	void ReleaseCallback_VoiceProcessingPassEnd(   PFUNC               pFunc ){ Callback_VoiceProcessingPassEnds.erase(   std::find( Callback_VoiceProcessingPassEnds.begin(),   Callback_VoiceProcessingPassEnds.end(),   pFunc ) ); };
	void ReleaseCallback_VoiceProcessingPassStart( PFUNC               pFunc ){ Callback_VoiceProcessingPassStarts.erase( std::find( Callback_VoiceProcessingPassStarts.begin(), Callback_VoiceProcessingPassStarts.end(), pFunc ) ); };
	void ReleaseCallback_BufferEnd(                PFUNC_VOIDP         pFunc ){ Callback_BufferEnds.erase(                std::find( Callback_BufferEnds.begin(),                Callback_BufferEnds.end(),                pFunc ) ); };
	void ReleaseCallback_BufferStart(              PFUNC_VOIDP         pFunc ){ Callback_BufferStarts.erase(              std::find( Callback_BufferStarts.begin(),              Callback_BufferStarts.end(),              pFunc ) ); };
	void ReleaseCallback_LoopEnd(                  PFUNC_VOIDP         pFunc ){ Callback_LoopEnds.erase(                  std::find( Callback_LoopEnds.begin(),                  Callback_LoopEnds.end(),                  pFunc ) ); };
	void ReleaseCallback_VoiceError(               PFUNC_VOIDP_HRESULT pFunc ){ Callback_VoiceErrors.erase(               std::find( Callback_VoiceErrors.begin(),               Callback_VoiceErrors.end(),               pFunc ) ); };

private:
	void Read( char *pBuffer, const unsigned int BufferSize, unsigned int &ReadSize, bool &IsEnd );

	// callback functions
	friend class VorbisCallback;

	void OnStreamEnd(){											for( std::vector < PFUNC               >::iterator it = Callback_StreamEnds.begin();                it != Callback_StreamEnds.end();                ++it ) (*it)(); };
	void OnVoiceProcessingPassEnd(){							for( std::vector < PFUNC               >::iterator it = Callback_VoiceProcessingPassEnds.begin();   it != Callback_VoiceProcessingPassEnds.end();   ++it ) (*it)(); };
	void OnVoiceProcessingPassStart(){							for( std::vector < PFUNC               >::iterator it = Callback_VoiceProcessingPassStarts.begin(); it != Callback_VoiceProcessingPassStarts.end(); ++it ) (*it)(); };
	void OnBufferEnd( void * pBufferContext ){					for( std::vector < PFUNC_VOIDP         >::iterator it = Callback_BufferEnds.begin();                it != Callback_BufferEnds.end();                ++it ) (*it)( pBufferContext ); };
	void OnBufferStart( void * pBufferContext ){				for( std::vector < PFUNC_VOIDP         >::iterator it = Callback_BufferStarts.begin();              it != Callback_BufferStarts.end();              ++it ) (*it)( pBufferContext ); };
	void OnLoopEnd( void * pBufferContext ){					for( std::vector < PFUNC_VOIDP         >::iterator it = Callback_LoopEnds.begin();                  it != Callback_LoopEnds.end();                  ++it ) (*it)( pBufferContext ); };
	void OnVoiceError( void * pBufferContext, HRESULT Error ){	for( std::vector < PFUNC_VOIDP_HRESULT >::iterator it = Callback_VoiceErrors.begin();               it != Callback_VoiceErrors.end();               ++it ) (*it)( pBufferContext, Error ); };

private:
	char* pBuffer[WAV_NUMBUFFER];
	int ValidBufferSize[WAV_NUMBUFFER];
	int BufferTopPos;

	OggVorbis_File vf;
	FILE *fp;

	bool IsBufferedAll;

	int NumLoops;

	XAudioFadeEffect *pXAudioFadeEffect;

	IXAudio2SourceVoice* pSourceVoice;
	VorbisCallback *pStatusCallBack;

	STATUS Status;

	std::vector < PFUNC >				Callback_StreamEnds;
	std::vector < PFUNC >				Callback_VoiceProcessingPassEnds;
	std::vector < PFUNC >				Callback_VoiceProcessingPassStarts;
	std::vector < PFUNC_VOIDP >			Callback_BufferEnds;
	std::vector < PFUNC_VOIDP >			Callback_BufferStarts;
	std::vector < PFUNC_VOIDP >			Callback_LoopEnds;
	std::vector < PFUNC_VOIDP_HRESULT >	Callback_VoiceErrors;
};

// Callback Function
class VorbisCallback : public IXAudio2VoiceCallback
{
public:
    VorbisCallback(): pXAudioVorbis( NULL ){}
    ~VorbisCallback(){}

    //Called when the voice has just finished playing a contiguous audio stream.
	STDMETHOD_( void, OnStreamEnd )					(){ pXAudioVorbis->OnStreamEnd(); };
    STDMETHOD_( void, OnVoiceProcessingPassEnd )	(){ pXAudioVorbis->OnVoiceProcessingPassEnd(); };
    STDMETHOD_( void, OnVoiceProcessingPassStart )	( UINT32 SamplesRequired ){ pXAudioVorbis->OnVoiceProcessingPassStart(); };
    STDMETHOD_( void, OnBufferEnd )					( void * pBufferContext ){ pXAudioVorbis->OnBufferEnd( pBufferContext ); };
    STDMETHOD_( void, OnBufferStart )				( void * pBufferContext ){ pXAudioVorbis->OnBufferStart( pBufferContext ); };
    STDMETHOD_( void, OnLoopEnd )					( void * pBufferContext ){ pXAudioVorbis->OnLoopEnd( pBufferContext ); };
    STDMETHOD_( void, OnVoiceError )				( void * pBufferContext, HRESULT Error ){ pXAudioVorbis->OnVoiceError( pBufferContext, Error ); };

public:
	XAudioVorbis *pXAudioVorbis;
};

#endif //__XAUDIOVORBIS_H__
