#include "XAudioWave.h"

#pragma warning( disable: 4996 )
#include <stdio.h>

XAudioWave::XAudioWave()
{
	pSourceVoice = NULL;
	WaveSize = 0;
	pData = NULL;
	pXAudioFadeEffect = new XAudioFadeEffect();
	pStatusCallBack = new WaveCallback();

	Status = STATUS_WAITING;
}

XAudioWave::~XAudioWave()
{
//	if( pXAudioFadeEffect ){
//		pXAudioFadeEffect->Release();
//	}

	Release();

	if( pStatusCallBack ){
		delete pStatusCallBack;
		pStatusCallBack = NULL;
	}

	if( pXAudioFadeEffect ){
		delete pXAudioFadeEffect;
		pXAudioFadeEffect = NULL;
	}
}

bool XAudioWave::Load( const char *pFileName )
{
	Stop();
	Release();

	FILE *fp = fopen( pFileName, "rb" );
	if( !fp ) return false;

	unsigned char Header[4];
	fread( Header, 4, 1, fp );

	// RIFF
	if( memcmp( Header, "RIFF", 4 ) != 0 ){
		fclose( fp );
		return false;
	}

	// File Size
	fread( Header, 4, 1, fp );
	int FileSize = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// WAVE
	fread( Header, 4, 1, fp );
	if( memcmp( Header, "WAVE", 4 ) != 0 ){
		fclose( fp );
		return false;
	}

	// fmt
	fread( Header, 4, 1, fp );
	if( memcmp( Header, "fmt ", 4 ) != 0 ){
		fclose( fp );
		return false;
	}

	// Linear PCM
	fread( Header, 4, 1, fp );
	int FMTChunkBytes = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// http://www.web-sky.org/program/other/wave.php
	// WAVE_FORMAT_UNKNOWN	0x0000	Microsoft Corporation
	// WAVE_FORMAT_PCM	0x0001	Microsoft PCM format
	// WAVE_FORMAT_MS_ADPCM	0x0002	Microsoft ADPCM
	// WAVE_FORMAT_IEEE_FLOAT	0x0003	Micrososft 32 bit float format
	// WAVE_FORMAT_VSELP	0x0004	Compaq Computer Corporation
	// WAVE_FORMAT_IBM_CVSD	0x0005	IBM Corporation
	// WAVE_FORMAT_ALAW	0x0006	Microsoft Corporation
	// WAVE_FORMAT_MULAW	0x0007	Microsoft Corporation
	// WAVE_FORMAT_OKI_ADPCM	0x0010	OKI
	// WAVE_FORMAT_IMA_ADPCM	0x0011	Intel Corporation
	// WAVE_FORMAT_MEDIASPACE_ADPCM	0x0012	Videologic
	// WAVE_FORMAT_SIERRA_ADPCM	0x0013	Sierra Semiconductor Corp
	// WAVE_FORMAT_G723_ADPCM	0x0014	Antex Electronics Corporation
	// WAVE_FORMAT_DIGISTD	0x0015	DSP Solutions, Inc.
	// WAVE_FORMAT_DIGIFIX	0x0016	DSP Solutions, Inc.
	// WAVE_FORMAT_DIALOGIC_OKI_ADPCM	0x0017	Dialogic Corporation
	// WAVE_FORMAT_MEDIAVISION_ADPCM	0x0018	Media Vision, Inc.
	// WAVE_FORMAT_CU_CODEC	0x0019	Hewlett-Packard Company
	// WAVE_FORMAT_YAMAHA_ADPCM	0x0020	Yamaha Corporation of America
	// WAVE_FORMAT_SONARC	0x0021	Speech Compression
	// WAVE_FORMAT_DSPGROUP_TRUESPEECH	0x0022	DSP Group, Inc
	// WAVE_FORMAT_ECHOSC1	0x0023	Echo Speech Corporation
	// WAVE_FORMAT_AUDIOFILE_AF36	0x0024	Audiofile, Inc.
	// WAVE_FORMAT_APTX	0x0025	Audio Processing Technology
	// WAVE_FORMAT_AUDIOFILE_AF10	0x0026	Audiofile, Inc.
	// WAVE_FORMAT_PROSODY_1612	0x0027	Aculab plc
	// WAVE_FORMAT_LRC	0x0028	Merging Technologies S.A.
	// WAVE_FORMAT_DOLBY_AC2	0x0030	Dolby Laboratories
	// WAVE_FORMAT_GSM610	0x0031	Microsoft Corporation
	// WAVE_FORMAT_MSNAUDIO	0x0032	Microsoft Corporation
	// WAVE_FORMAT_ANTEX_ADPCME	0x0033	Antex Electronics Corporation
	// WAVE_FORMAT_CONTROL_RES_VQLPC	0x0034	Control Resources Limited
	// WAVE_FORMAT_DIGIREAL	0x0035	DSP Solutions, Inc.
	// WAVE_FORMAT_DIGIADPCM	0x0036	DSP Solutions, Inc.
	// WAVE_FORMAT_CONTROL_RES_CR10	0x0037	Control Resources Limited
	// WAVE_FORMAT_NMS_VBXADPCM	0x0038	Natural MicroSystems
	// WAVE_FORMAT_ROLAND_RDAC	0x0039	Roland
	// WAVE_FORMAT_ECHOSC3	0x003A	Echo Speech Corporation
	// WAVE_FORMAT_ROCKWELL_ADPCM	0x003B	Rockwell International
	// WAVE_FORMAT_ROCKWELL_DIGITALK	0x003C	Rockwell International
	// WAVE_FORMAT_XEBEC	0x003D	Xebec Multimedia Solutions Limited
	// WAVE_FORMAT_G721_ADPCM	0x0040	Antex Electronics Corporation
	// WAVE_FORMAT_G728_CELP	0x0041	Antex Electronics Corporation
	// WAVE_FORMAT_MSG723	0x0042	Microsoft Corporation
	// WAVE_FORMAT_MPEG	0x0050	Microsoft Corporation
	// WAVE_FORMAT_RT24	0x0052	InSoft Inc.
	// WAVE_FORMAT_PAC	0x0053	InSoft Inc.
	// WAVE_FORMAT_MPEGLAYER3	0x0055	MPEG 3 Layer 1
	// WAVE_FORMAT_LUCENT_G723	0x0059	Lucent Technologies
	// WAVE_FORMAT_CIRRUS	0x0060	Cirrus Logic
	// WAVE_FORMAT_ESPCM	0x0061	ESS Technology
	// WAVE_FORMAT_VOXWARE	0x0062	Voxware Inc
	// WAVE_FORMAT_CANOPUS_ATRAC	0x0063	Canopus, Co., Ltd.
	// WAVE_FORMAT_G726_ADPCM	0x0064	APICOM
	// WAVE_FORMAT_G722_ADPCM	0x0065	APICOM
	// WAVE_FORMAT_DSAT	0x0066	Microsoft Corporation
	// WAVE_FORMAT_DSAT_DISPLAY	0x0067	Microsoft Corporation
	// WAVE_FORMAT_VOXWARE_BYTE_ALIGNED	0x0069	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_AC8	0x0070	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_AC10	0x0071	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_AC16	0x0072	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_AC20	0x0073	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_RT24	0x0074	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_RT29	0x0075	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_RT29HW	0x0076	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_VR12	0x0077	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_VR18	0x0078	Voxware Inc.
	// WAVE_FORMAT_VOXWARE_TQ40	0x0079	Voxware Inc.
	// WAVE_FORMAT_SOFTSOUND	0x0080	Softsound, Ltd.
	// WAVE_FORMAT_VOXARE_TQ60	0x0081	Voxware Inc.
	// WAVE_FORMAT_MSRT24	0x0082	Microsoft Corporation
	// WAVE_FORMAT_G729A	0x0083	AT&T Laboratories
	// WAVE_FORMAT_MVI_MV12	0x0084	Motion Pixels
	// WAVE_FORMAT_DF_G726	0x0085	DataFusion Systems (Pty) (Ltd)
	// WAVE_FORMAT_DF_GSM610	0x0086	DataFusion Systems (Pty) (Ltd)
	// WAVE_FORMAT_ONLIVE	0x0089	OnLive! Technologies, Inc.
	// WAVE_FORMAT_SBC24	0x0091	Siemens Business Communications Systems
	// WAVE_FORMAT_DOLBY_AC3_SPDIF	0x0092	Sonic Foundry
	// WAVE_FORMAT_ZYXEL_ADPCM	0x0097	ZyXEL Communications, Inc.
	// WAVE_FORMAT_PHILIPS_LPCBB	0x0098	Philips Speech Processing
	// WAVE_FORMAT_PACKED	0x0099	Studer Professional Audio AG
	// WAVE_FORMAT_RHETOREX_ADPCM	0x0100	Rhetorex, Inc.
	// IBM_FORMAT_MULAW	0x0101	IBM mu-law format
	// IBM_FORMAT_ALAW	0x0102	IBM a-law format
	// IBM_FORMAT_ADPCM	0x0103	IBM AVC Adaptive Differential PCM format
	// WAVE_FORMAT_VIVO_G723	0x0111	Vivo Software
	// WAVE_FORMAT_VIVO_SIREN	0x0112	Vivo Software
	// WAVE_FORMAT_DIGITAL_G723	0x0123	Digital Equipment Corporation
	// WAVE_FORMAT_CREATIVE_ADPCM	0x0200	Creative Labs, Inc
	// WAVE_FORMAT_CREATIVE_FASTSPEECH8	0x0202	Creative Labs, Inc
	// WAVE_FORMAT_CREATIVE_FASTSPEECH10	0x0203	Creative Labs, Inc
	// WAVE_FORMAT_QUARTERDECK	0x0220	Quarterdeck Corporation
	// WAVE_FORMAT_FM_TOWNS_SND	0x0300	Fujitsu Corporation
	// WAVE_FORMAT_BZV_DIGITAL	0x0400	Brooktree Corporation
	// WAVE_FORMAT_VME_VMPCM	0x0680	AT&T Labs, Inc.
	// WAVE_FORMAT_OLIGSM	0x1000	Ing C. Olivetti & C., S.p.A.
	// WAVE_FORMAT_OLIADPCM	0x1001	Ing C. Olivetti & C., S.p.A.
	// WAVE_FORMAT_OLICELP	0x1002	Ing C. Olivetti & C., S.p.A.
	// WAVE_FORMAT_OLISBC	0x1003	Ing C. Olivetti & C., S.p.A.
	// WAVE_FORMAT_OLIOPR	0x1004	Ing C. Olivetti & C., S.p.A.
	// WAVE_FORMAT_LH_CODEC	0x1100	Lernout & Hauspie
	// WAVE_FORMAT_NORRIS	0x1400	Norris Communications, Inc.
	// WAVE_FORMAT_SOUNDSPACE_MUSICOMPRESS	0x1500	AT&T Labs, Inc.
	// WAVE_FORMAT_DVM	0x2000	FAST Multimedia AG
	// WAVE_FORMAT_INTERWAV_VSC112	0x7150	?????
	// WAVE_FORMAT_EXTENSIBLE	0xFFFE

	fread( Header, 2, 1, fp );
	int FormatID = ( ( (int)Header[1] ) << 8 ) + (int)Header[0];
	if( FormatID != 0x0001 && FormatID != 0xFFFE ){ // 0x0001: WAVE_FORMAT_PCM 0xFFFE: WAVE_FORMAT_EXTENSIBLE
		fclose( fp );
		return false;
	}

	// Number of Channels
	fread( Header, 2, 1, fp );
	int NumChannels = ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// Sampling Rate
	fread( Header, 4, 1, fp );
	int SamplingRate = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// Data Speed
	fread( Header, 4, 1, fp );
	int DataSpeed = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// Block Size
	fread( Header, 2, 1, fp );
	int BlockSize = ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// Bits / Sample
	fread( Header, 2, 1, fp );
	int BitsPerSample = ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	// Skip
	if( FMTChunkBytes - 16 > 0 ) fseek( fp, FMTChunkBytes - 16, SEEK_CUR );

	// skip chunks
	for( int i = 0; ; ++i ){
		fread( Header, 4, 1, fp );

		if( memcmp( Header, "data", 4 ) == 0 ){
			break;
		}

		fread( Header, 4, 1, fp );

		int Size = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];
		fseek( fp, Size, SEEK_CUR );
	}

	// data
	// WaveSize (byte)
	fread( Header, 4, 1, fp );
	WaveSize = ( ( (int)Header[3] ) << 24 ) + ( ( (int)Header[2] ) << 16 ) + ( ( (int)Header[1] ) << 8 ) + (int)Header[0];

	if( WaveSize > XAUDIO2_MAX_BUFFER_BYTES ) {
		fclose( fp );
		return false;
	}

	pData = new unsigned char[WaveSize];

	fread( pData, sizeof( char ), WaveSize, fp );

	fclose( fp );

	// PCM audio
	WAVEFORMATEX WaveFormatEx;

	WaveFormatEx.wFormatTag = 1; // PCM audio
	WaveFormatEx.nChannels = NumChannels;
	WaveFormatEx.nSamplesPerSec = SamplingRate;
	WaveFormatEx.wBitsPerSample = BitsPerSample;
	WaveFormatEx.nBlockAlign = WaveFormatEx.nChannels * WaveFormatEx.wBitsPerSample / 8;
	WaveFormatEx.nAvgBytesPerSec = WaveFormatEx.nSamplesPerSec * WaveFormatEx.nBlockAlign;
	WaveFormatEx.cbSize = 0;

	pStatusCallBack->pXAudioWave = this;

	if( FAILED( XAudio::GetInstance().GetXAudio2()->CreateSourceVoice(
			&pSourceVoice, &WaveFormatEx, 0, XAUDIO2_DEFAULT_FREQ_RATIO, pStatusCallBack ) ) ){
		return false;
	}

	if( !pXAudioFadeEffect->Init( pSourceVoice, NumChannels ) ) return false;


	FadeEffectParams Param;
	Param.IsFade = false;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	SetCallback_BufferEnd( Callback_BufferEnd );

	return true;
}

bool XAudioWave::Release()
{
	Stop();

	if( pData ){
		delete [] pData;
		pData = NULL;
	}

	if( pSourceVoice ){
		pSourceVoice->FlushSourceBuffers();

		for( int i = 0; i < 10; ++i ){ // 10 ͓K
	        XAUDIO2_VOICE_STATE s;
			pSourceVoice->GetState( &s );
			if( s.BuffersQueued == 0 ) break;

			Sleep( 100 );
		}

        pSourceVoice->DestroyVoice();
		pSourceVoice = NULL;

		ReleaseCallback_BufferEnd( Callback_BufferEnd );
	}

	Status = STATUS_WAITING;

	return true;
}

bool XAudioWave::Play( int NumLoops )
{
	if( pSourceVoice ){
		Stop();
		Fade( 0, 1.0f, 1.0f );

		// Submit the wave sample data using an XAUDIO2_BUFFER structure
		XAUDIO2_BUFFER XAudio2Buffer;
		memset( &XAudio2Buffer, 0, sizeof( XAUDIO2_BUFFER ) );

		XAudio2Buffer.AudioBytes = WaveSize;
		XAudio2Buffer.pAudioData = pData;
		XAudio2Buffer.pContext = this;
		XAudio2Buffer.Flags = XAUDIO2_END_OF_STREAM;

		if( NumLoops >= 0 ){
			XAudio2Buffer.LoopCount = NumLoops;
		}
		else{
			XAudio2Buffer.LoopCount = XAUDIO2_LOOP_INFINITE;
		}

		if( FAILED( pSourceVoice->SubmitSourceBuffer( &XAudio2Buffer ) ) ){
			pSourceVoice->DestroyVoice();
			pSourceVoice = NULL;
			return false;
		}

		if( FAILED( pSourceVoice->Start( 0 ) ) ){
			return false;
		}
	}

	Status = STATUS_PLAYING;

	return true;
}

bool XAudioWave::Resume()
{
	if( pSourceVoice ){
		if( FAILED( pSourceVoice->Start( 0 ) ) ){
			return false;
		}
	}

	Status = STATUS_PLAYING;

	return true;
}

bool XAudioWave::Stop()
{
	if( pSourceVoice ){
		if( FAILED( pSourceVoice->Stop( XAUDIO2_PLAY_TAILS ) ) ){
			return false;
		}
	}

	Status = STATUS_STOP;

	return true;
}

// -2^24 to 2^24
// 1.0 means there is no attenuation or gain and 0 means silence. Negative levels can be used to invert the audio's phase.
bool XAudioWave::SetVolume( const float Volume )
{
	if( Volume < -16777216.0 || Volume > 16777216.0 ) return false;

	if( FAILED( pSourceVoice->SetVolume( Volume ) ) ){
		return false;
	}

	return true;
}

float XAudioWave::GetVolume() const
{
	float Volume;
	pSourceVoice->GetVolume( &Volume );

	return Volume;
}

bool XAudioWave::SetFrequencyRatio( const float Ratio )
{
	if( Ratio < XAUDIO2_MIN_FREQ_RATIO || Ratio > XAUDIO2_DEFAULT_FREQ_RATIO ) return false;

	if( FAILED( pSourceVoice->SetFrequencyRatio( Ratio ) ) ){
		return false;
	}

	return true;
}

float XAudioWave::GetFrequencyRatio() const
{
	float Ratio;
	pSourceVoice->GetFrequencyRatio( &Ratio );

	return Ratio;
}

bool XAudioWave::Fade( const unsigned int Milliseconds, const float BeginGain, const float EndGain )
{
	FadeEffectParams Param;
	Param.IsFade = true;
	Param.BeginGain = BeginGain;
	Param.EndGain = EndGain;
	Param.Milliseconds = Milliseconds;
	Param.Frames = 0;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	return true;
}

// PrevParam.Gain̏KvȂ̂ŁAKXAudioWave::Fade( unsigned int Milliseconds, const float BeginGain, const float EndGain )Ă񂾌ŎgƁB
bool XAudioWave::Fade( const unsigned int Milliseconds, const float EndGain )
{
	FadeEffectParams PrevParam;
	if( FAILED( pSourceVoice->GetEffectParameters( 0, &PrevParam, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	FadeEffectParams Param;
	Param.IsFade = true;
	Param.BeginGain = PrevParam.Gain;
	Param.EndGain = EndGain;
	Param.Milliseconds = Milliseconds;
	Param.Frames = 0;

	if( FAILED( pSourceVoice->SetEffectParameters( 0, &Param, sizeof( FadeEffectParams ) ) ) ){
		return false;
	}

	return true;
}

void XAudioWave::Callback_BufferEnd( void *p )
{
	XAudioWave *pXAudioWave = (XAudioWave *)p;

	pXAudioWave->Status = STATUS_WAITING;
}
